@@ -4,13 +4,15 @@ use crate::import::sheet::TilesetGenInfo;
 
 use keeshond::gameloop::GameControl;
 use keeshond::renderer::{Sheet, SpriteSlice};
-use keeshond_datapack::DataId;
+use keeshond::datapack::DataId;
 
 use std::any::Any;
 
 use image::{RgbaImage, GrayImage};
+use crate::util::zoom::ZoomControl;
 
 const MAX_DRAW_POINTS : usize = 32768;
+const ZOOM_LEVELS : [f32 ; 11] = [0.25, 0.5, 1.0, 2.0, 3.0, 4.0, 6.0, 8.0, 10.0, 12.0, 16.0];
 
 enum SheetAction
 {
@@ -32,7 +34,8 @@ pub struct SheetDocument
     data_id : DataId<Sheet>,
     show_origin : bool,
     transparent_color : [f32; 3],
-    zoom_level : f32,
+    zoom_control : ZoomControl<f32>,
+    dpi_factor : f32,
     current_slice : usize,
     tics : u64,
     draw_tool_on : bool,
@@ -41,6 +44,9 @@ pub struct SheetDocument
     scroll_slice_list : bool,
     mouse_drawing_texture : bool,
     draw_points : Vec<(f32, f32)>,
+    sheet_scroll_x : f32,
+    sheet_scroll_y : f32,
+    update_sheet_scroll : bool,
     dragging : bool,
     tilegen : TilesetGenInfo,
     tilegen_preview : bool,
@@ -49,6 +55,21 @@ pub struct SheetDocument
 
 impl SheetDocument
 {
+    fn adjusted_zoom(&self) -> f32
+    {
+        self.zoom_control.zoom() / self.dpi_factor
+    }
+
+    fn scaled_floor(&self, value : f32) -> f32
+    {
+        crate::util::scaled_floor(value, 1.0 / self.dpi_factor)
+    }
+
+    fn scaled_round(&self, value : f32) -> f32
+    {
+        crate::util::scaled_round(value, 1.0 / self.dpi_factor)
+    }
+
     fn opaque_pixels(&self, sheet : &Sheet) -> toodee::TooDee<bool>
     {
         let mut pixels = toodee::TooDee::new(sheet.width() as usize, sheet.height() as usize);
@@ -72,10 +93,10 @@ impl SheetDocument
 
     fn draw_slice_bounds(&mut self, pos: [f32; 2], draw_list: &imgui::DrawListMut, slice: &SpriteSlice, active : bool)
     {
-        let mut x = slice.texture_x * self.zoom_level + pos[0];
-        let mut y = slice.texture_y * self.zoom_level + pos[1];
-        let mut x2 = x + slice.texture_w * self.zoom_level;
-        let mut y2 = y + slice.texture_h * self.zoom_level;
+        let mut x = self.scaled_round(slice.texture_x * self.adjusted_zoom() + pos[0]);
+        let mut y = self.scaled_round(slice.texture_y * self.adjusted_zoom() + pos[1]);
+        let mut x2 = self.scaled_round(x + slice.texture_w * self.adjusted_zoom());
+        let mut y2 = self.scaled_round(y + slice.texture_h * self.adjusted_zoom());
 
         let color;
         let thickness;
@@ -89,8 +110,8 @@ impl SheetDocument
 
             let pulse = (((self.tics as f64 * 0.1).sin() + 1.0) / 2.0) as f32;
 
-            let origin_x = x + (slice.origin_x * self.zoom_level) + 1.0;
-            let origin_y = y + (slice.origin_y * self.zoom_level) + 1.0;
+            let origin_x = x + (slice.origin_x * self.adjusted_zoom()) + 1.0;
+            let origin_y = y + (slice.origin_y * self.adjusted_zoom()) + 1.0;
 
             color = [1.0, pulse, pulse, 1.0];
             thickness = ((self.tics as f64 * 0.05).sin() + 1.0) as f32;
@@ -106,18 +127,18 @@ impl SheetDocument
             color = [1.0, 1.0, 1.0, 1.0];
         }
 
-        draw_list.add_rect([x, y], [x2, y2], imgui::ImColor32::from(color)).thickness(thickness).build();
+        draw_list.add_rect([x, y], [x2, y2], imgui::ImColor32::from(color)).thickness(thickness / self.dpi_factor).build();
     }
 
     fn draw_crosshair(&mut self, draw_list : &imgui::DrawListMut, x : f32, y : f32)
     {
         let pulse = (((self.tics as f64 * 0.1).sin() + 1.0) / 2.0) as f32;
 
-        let cross_size = 4.1 + self.zoom_level.min(8.0);
+        let cross_size = 4.1 + self.adjusted_zoom().min(8.0);
         let cross_color = [pulse, pulse, pulse, 1.0];
 
-        draw_list.add_line([x - cross_size, y], [x + cross_size, y], imgui::ImColor32::from(cross_color)).thickness(1.0).build();
-        draw_list.add_line([x, y - cross_size], [x, y + cross_size], imgui::ImColor32::from(cross_color)).thickness(1.0).build();
+        draw_list.add_line([x - cross_size, y], [x + cross_size, y], imgui::ImColor32::from(cross_color)).thickness(1.0 / self.dpi_factor).build();
+        draw_list.add_line([x, y - cross_size], [x, y + cross_size], imgui::ImColor32::from(cross_color)).thickness(1.0 / self.dpi_factor).build();
     }
 
     fn defringe_sheet(&mut self, sheet : &mut Sheet)
@@ -125,7 +146,7 @@ impl SheetDocument
         let mut pixels = sheet.data();
         let mut texture = RgbaImage::from_raw(sheet.width(), sheet.height(), pixels).expect("Raw data to image failed");
 
-        crate::util::defringe_image(&mut texture, true);
+        crate::util::sheet::defringe_image(&mut texture, true);
 
         pixels = texture.into_raw();
 
@@ -290,7 +311,7 @@ impl SheetDocument
             return;
         }
 
-        crate::util::flood_fill(&mut pixels, x, y, true, |tiles, x, y|
+        crate::util::sheet::flood_fill(&mut pixels, x, y, true, |tiles, x, y|
         {
             tiles[(x, y)] == true
         },
@@ -307,7 +328,8 @@ impl SheetDocument
 
     fn draw_tileset_preview(&self, pos: [f32; 2], draw_list : &imgui::DrawListMut, sheet : &Sheet)
     {
-        if self.tilegen.width as f32 * self.zoom_level < 8.0 || self.tilegen.height as f32 * self.zoom_level < 8.0
+        if self.tilegen.width as f32 * self.adjusted_zoom() < 8.0
+            || self.tilegen.height as f32 * self.adjusted_zoom() < 8.0
         {
             return;
         }
@@ -322,12 +344,12 @@ impl SheetDocument
         {
             for x in 0..tileset_info.columns
             {
-                let x1 = pos[0] + (self.tilegen.x as f32 + (tileset_info.tile_walk_x * x) as f32) * self.zoom_level;
-                let y1 = pos[1] + (self.tilegen.y as f32 + (tileset_info.tile_walk_y * y) as f32) * self.zoom_level;
-                let x2 = x1 + (self.tilegen.width as f32) * self.zoom_level;
-                let y2 = y1 + (self.tilegen.height as f32) * self.zoom_level;
+                let x1 = pos[0] + (self.tilegen.x as f32 + (tileset_info.tile_walk_x * x) as f32) * self.adjusted_zoom();
+                let y1 = pos[1] + (self.tilegen.y as f32 + (tileset_info.tile_walk_y * y) as f32) * self.adjusted_zoom();
+                let x2 = x1 + (self.tilegen.width as f32) * self.adjusted_zoom();
+                let y2 = y1 + (self.tilegen.height as f32) * self.adjusted_zoom();
 
-                draw_list.add_rect([x1, y1], [x2, y2], imgui::ImColor32::from(color)).thickness(thickness).build();
+                draw_list.add_rect([x1, y1], [x2, y2], imgui::ImColor32::from(color)).thickness(thickness / self.dpi_factor).build();
             }
         }
     }
@@ -363,6 +385,12 @@ impl SheetDocument
         }
     }
 
+    fn apply_zoom(&mut self, level : f32)
+    {
+        self.zoom_control.set_zoom(level);
+        self.update_sheet_scroll = true;
+    }
+
     fn sheet_tab(&mut self, ui : &imgui::Ui, sheet : &mut Sheet)
     {
         let (mut force, mut force_on) = crate::util::option_bool_to_double_bool(sheet.texture_filtering());
@@ -544,8 +572,6 @@ impl SheetDocument
         style_slice_list.pop();
     }
 
-
-
     fn tilegen_tab(&mut self, ui : &imgui::Ui, sheet : &mut Sheet)
     {
         self.tilegen.tilegen_ui(ui);
@@ -591,139 +617,187 @@ impl SheetDocument
     {
         let style_var_outer = ui.push_style_var(imgui::StyleVar::WindowPadding([0.0, 0.0]));
 
-        imgui::ChildWindow::new("texture_view").size([0.0, 0.0]).draw_background(true).border(true).horizontal_scrollbar(true).movable(false).build(ui, ||
+        imgui::ChildWindow::new("texture_view").flags(imgui::WindowFlags::NO_SCROLL_WITH_MOUSE)
+            .size([0.0, 0.0]).draw_background(true).border(true)
+            .horizontal_scrollbar(true).movable(false).build(ui, ||
         {
+            let window_pos = ui.window_pos();
+            let size = ui.window_size();
+
+            let mut real_scroll_x = ui.scroll_x();
+            let mut real_scroll_y = ui.scroll_y();
+
+            if ui.is_window_hovered_with_flags(imgui::WindowHoveredFlags::CHILD_WINDOWS)
+            {
+                if self.zoom_control.update_mouse_wheel(ui.io().mouse_wheel)
+                {
+                    self.apply_zoom(self.zoom_control.zoom());
+                }
+            }
+
+            let mut mouse_x_ratio = 0.5;
+            let mut mouse_y_ratio = 0.5;
+
+            if ui.is_window_hovered_with_flags(imgui::WindowHoveredFlags::CHILD_WINDOWS)
+            {
+                mouse_x_ratio = ((ui.io().mouse_pos[0] - window_pos[0]) / size[0]).clamp(0.0, 1.0);
+                mouse_y_ratio = ((ui.io().mouse_pos[1] - window_pos[1]) / size[1]).clamp(0.0, 1.0);
+            }
+
+            if self.update_sheet_scroll
+            {
+                self.update_sheet_scroll = false;
+
+                let real_scroll_x_max = (sheet.width() as f32 * self.adjusted_zoom() - size[0]).max(0.0);
+                let real_scroll_y_max = (sheet.height() as f32 * self.adjusted_zoom() - size[1]).max(0.0);
+
+                real_scroll_x = (self.sheet_scroll_x * self.adjusted_zoom() - size[0] * mouse_x_ratio)
+                    .clamp(0.0, real_scroll_x_max);
+                real_scroll_y = (self.sheet_scroll_y * self.adjusted_zoom() - size[1] * mouse_y_ratio)
+                    .clamp(0.0, real_scroll_y_max);
+                ui.set_scroll_x(real_scroll_x);
+                ui.set_scroll_y(real_scroll_y);
+            }
+
+            self.sheet_scroll_x = (real_scroll_x + size[0] * mouse_x_ratio + 0.5) / self.adjusted_zoom();
+            self.sheet_scroll_y = (real_scroll_y + size[1] * mouse_y_ratio + 0.5) / self.adjusted_zoom();
+
             let mut delta = [0.0, 0.0];
 
-            let style_var = ui.push_style_var(imgui::StyleVar::WindowPadding([1.0, 1.0]));
-            let style_color = ui.push_style_color(imgui::StyleColor::ChildBg, [self.transparent_color[0], self.transparent_color[1], self.transparent_color[2], 1.0]);
+            let width = (sheet.width() as f32 * self.adjusted_zoom()).max(1.0);
+            let height = (sheet.height() as f32 * self.adjusted_zoom()).max(1.0);
 
-            let width = (sheet.width() as f32 * self.zoom_level).max(1.0);
-            let height = (sheet.height() as f32 * self.zoom_level).max(1.0);
+            let pos_min = [
+                self.scaled_round(window_pos[0] - real_scroll_x),
+                self.scaled_round(window_pos[1] - real_scroll_y)];
+            let pos_max = [
+                self.scaled_round(pos_min[0] + width),
+                self.scaled_round(pos_min[1] + height)];
 
-            imgui::ChildWindow::new("texture_item").size([width + 2.0, height + 2.0])
-                .scrollable(false).draw_background(true).border(true).build(ui, ||
+            ui.invisible_button("texture_drag", [width, height]);
+
+            let hovering = ui.is_item_hovered();
+
+            if ui.is_mouse_clicked(imgui::MouseButton::Left) && hovering && !self.on_tilegen
             {
-                let pos = ui.cursor_screen_pos();
-
-                ui.invisible_button("texture_drag", [width, height]);
-
-                let hovering = ui.is_item_hovered();
-
-                if ui.is_mouse_clicked(imgui::MouseButton::Left) && hovering && !self.on_tilegen
+                if self.draw_tool_on
                 {
-                    if self.draw_tool_on
-                    {
-                        self.mouse_drawing_texture = true;
-                    }
-                    else
-                    {
-                        let mouse_pos = ui.io().mouse_pos;
-                        let new_point = (((mouse_pos[0] - pos[0]) / self.zoom_level).floor(), ((mouse_pos[1] - pos[1]) / self.zoom_level).floor());
-
-                        self.select_slice(sheet, new_point);
-                    }
+                    self.mouse_drawing_texture = true;
                 }
-
-                if self.mouse_drawing_texture && ui.is_mouse_down(imgui::MouseButton::Left) && hovering
+                else
                 {
                     let mouse_pos = ui.io().mouse_pos;
-                    let new_point = (((mouse_pos[0] - pos[0]) / self.zoom_level).floor(), ((mouse_pos[1] - pos[1]) / self.zoom_level).floor());
-                    let mut push = true;
+                    let new_point = (
+                        self.scaled_floor((mouse_pos[0] - pos_min[0]) / self.adjusted_zoom()),
+                        self.scaled_floor((mouse_pos[1] - pos_min[1]) / self.adjusted_zoom())
+                    );
 
-                    if let Some(last) = self.draw_points.last()
+                    self.select_slice(sheet, new_point);
+                }
+            }
+
+            if self.mouse_drawing_texture && ui.is_mouse_down(imgui::MouseButton::Left) && hovering
+            {
+                let mouse_pos = ui.io().mouse_pos;
+                let new_point = (
+                    self.scaled_floor((mouse_pos[0] - pos_min[0]) / self.adjusted_zoom()),
+                    self.scaled_floor((mouse_pos[1] - pos_min[1]) / self.adjusted_zoom())
+                );
+                let mut push = true;
+
+                if let Some(last) = self.draw_points.last()
+                {
+                    push = *last != new_point;
+                }
+
+                if push
+                {
+                    self.draw_points.push(new_point);
+
+                    if self.draw_points.len() > MAX_DRAW_POINTS
                     {
-                        push = *last != new_point;
-                    }
-
-                    if push
-                    {
-                        self.draw_points.push(new_point);
-
-                        if self.draw_points.len() > MAX_DRAW_POINTS
-                        {
-                            self.draw_points.remove(0);
-                        }
+                        self.draw_points.remove(0);
                     }
                 }
-                else if !ui.is_mouse_down(imgui::MouseButton::Left)
-                {
-                    self.mouse_drawing_texture = false;
+            }
+            else if !ui.is_mouse_down(imgui::MouseButton::Left)
+            {
+                self.mouse_drawing_texture = false;
 
-                    if self.draw_points.len() > 0
+                if self.draw_points.len() > 0
+                {
+                    self.focus_slice_name_input = self.make_slice_from_draw_points(sheet);
+                }
+            }
+
+            if ui.is_mouse_clicked(imgui::MouseButton::Right) && hovering
+            {
+                self.dragging = true;
+            }
+
+            if self.dragging && ui.is_mouse_down(imgui::MouseButton::Right)
+            {
+                delta = ui.io().mouse_delta;
+            }
+            else if !ui.is_mouse_down(imgui::MouseButton::Right)
+            {
+                self.dragging = false;
+            }
+
+            ui.same_line_with_pos(1.0);
+            let draw_list = ui.get_window_draw_list();
+
+            draw_list.add_rect(pos_min, pos_max, [self.transparent_color[0], self.transparent_color[1], self.transparent_color[2], 1.0])
+                .filled(true).build();
+            draw_list.add_rect(pos_min, pos_max, ui.style_color(imgui::StyleColor::Border)).build();
+            draw_list.add_image(imgui::TextureId::from(self.data_id.index()), pos_min, pos_max).build();
+
+            if !self.on_tilegen
+            {
+                for i in 0..sheet.slice_count()
+                {
+                    if let Some(slice) = sheet.slice(i)
                     {
-                        self.focus_slice_name_input = self.make_slice_from_draw_points(sheet);
+                        self.draw_slice_bounds(pos_min, &draw_list, slice, false);
                     }
                 }
 
-                if ui.is_mouse_clicked(imgui::MouseButton::Right) && hovering
+                if let Some(slice) = sheet.slice(self.current_slice)
                 {
-                    self.dragging = true;
+                    self.draw_slice_bounds(pos_min, &draw_list, slice, true);
                 }
 
-                if self.dragging && ui.is_mouse_down(imgui::MouseButton::Right)
+                for point in 1..self.draw_points.len()
                 {
-                    delta = ui.io().mouse_delta;
-                }
-                else if !ui.is_mouse_down(imgui::MouseButton::Right)
-                {
-                    self.dragging = false;
-                }
+                    let (mut x1, mut y1) = self.draw_points[point - 1];
+                    let (mut x2, mut y2) = self.draw_points[point];
 
-                ui.same_line_with_pos(1.0);
-                imgui::Image::new(imgui::TextureId::from(self.data_id.index()), [width, height]).build(ui);
-
-                if !self.on_tilegen
-                {
-                    let draw_list = ui.get_window_draw_list();
-
-                    for i in 0..sheet.slice_count()
+                    if self.zoom_control.zoom() >= 3.0
                     {
-                        if let Some(slice) = sheet.slice(i)
-                        {
-                            self.draw_slice_bounds(pos, &draw_list, slice, false);
-                        }
+                        x1 += 0.5;
+                        y1 += 0.5;
+                        x2 += 0.5;
+                        y2 += 0.5;
                     }
 
-                    if let Some(slice) = sheet.slice(self.current_slice)
-                    {
-                        self.draw_slice_bounds(pos, &draw_list, slice, true);
-                    }
-
-                    for point in 1..self.draw_points.len()
-                    {
-                        let (mut x1, mut y1) = self.draw_points[point - 1];
-                        let (mut x2, mut y2) = self.draw_points[point];
-
-                        if self.zoom_level >= 3.0
-                        {
-                            x1 += 0.5;
-                            y1 += 0.5;
-                            x2 += 0.5;
-                            y2 += 0.5;
-                        }
-
-                        draw_list.add_line([x1 * self.zoom_level + pos[0], y1 * self.zoom_level + pos[1]],
-                                           [x2 * self.zoom_level + pos[0], y2 * self.zoom_level + pos[1]],
-                                           imgui::ImColor32::from([1.0, 1.0, 1.0, 1.0])).build();
-                    }
+                    draw_list.add_line([self.scaled_round(x1 * self.adjusted_zoom()) + pos_min[0],
+                                           self.scaled_round(y1 * self.adjusted_zoom()) + pos_min[1]],
+                                       [self.scaled_round(x2 * self.adjusted_zoom()) + pos_min[0],
+                                           self.scaled_round(y2 * self.adjusted_zoom()) + pos_min[1]],
+                                       imgui::ImColor32::from([1.0, 1.0, 1.0, 1.0]))
+                        .thickness(1.0 / self.dpi_factor).build();
                 }
-                else if self.tilegen_preview
-                {
-                    let draw_list = ui.get_window_draw_list();
-
-                    self.draw_tileset_preview(pos, &draw_list, sheet);
-                }
-            });
+            }
+            else if self.tilegen_preview
+            {
+                self.draw_tileset_preview(pos_min, &draw_list, sheet);
+            }
 
             if delta[0] != 0.0 || delta[1] != 0.0
             {
                 ui.set_scroll_x(ui.scroll_x() - delta[0]);
                 ui.set_scroll_y(ui.scroll_y() - delta[1]);
             }
-
-            style_color.pop();
-            style_var.pop();
         });
 
         style_var_outer.pop();
@@ -737,6 +811,14 @@ impl SheetDocument
         imgui::ChildWindow::new("texture_slice_item").size([0.0, 0.0])
             .scrollable(false).scroll_bar(false).draw_background(true).border(true).build(ui, ||
         {
+            if ui.is_window_hovered_with_flags(imgui::WindowHoveredFlags::CHILD_WINDOWS)
+            {
+                if self.zoom_control.update_mouse_wheel(ui.io().mouse_wheel)
+                {
+                    self.apply_zoom(self.zoom_control.zoom());
+                }
+            }
+
             if let Some(slice) = sheet.slice(self.current_slice)
             {
                 let base_width = sheet.width() as f32;
@@ -747,17 +829,21 @@ impl SheetDocument
                 let y2 = (slice.texture_y + slice.texture_h) / base_height;
 
                 let slice_space = ui.window_size();
-                let center_x = (slice_space[0] / 2.0).round();
-                let center_y = (slice_space[1] / 2.0).round();
-                let image_x = (center_x - slice.origin_x * self.zoom_level).round();
-                let image_y = (center_y - slice.origin_y * self.zoom_level).round();
+                let center_x = self.scaled_round(slice_space[0] / 2.0);
+                let center_y = self.scaled_round(slice_space[1] / 2.0);
+                let image_x = self.scaled_round(center_x - slice.origin_x * self.adjusted_zoom());
+                let image_y = self.scaled_round(center_y - slice.origin_y * self.adjusted_zoom());
 
                 ui.set_cursor_pos([image_x, image_y]);
+                crate::util::widget::round_cursor_screen_pos(ui);
 
-                imgui::Image::new(imgui::TextureId::from(self.data_id.index()), [slice.texture_w * self.zoom_level, slice.texture_h * self.zoom_level])
+                imgui::Image::new(imgui::TextureId::from(self.data_id.index()),
+                                  [self.scaled_round(slice.texture_w * self.adjusted_zoom()),
+                                      self.scaled_round(slice.texture_h * self.adjusted_zoom())])
                     .uv0([x1, y1]).uv1([x2, y2]).build(ui);
 
                 ui.set_cursor_pos([center_x, center_y]);
+                crate::util::widget::round_cursor_screen_pos(ui);
 
                 let draw_list = ui.get_window_draw_list();
                 let cursor_pos = ui.cursor_screen_pos();
@@ -789,13 +875,16 @@ impl Document for SheetDocument
             data_id = id;
         }
 
+        let dpi_factor = game.renderer().current_display_scale_factor();
+
         SheetDocument
         {
             action_history : ActionHistory::new(),
             data_id,
             show_origin : true,
             transparent_color : [0.5, 0.5001, 0.5002],
-            zoom_level : 1.0,
+            zoom_control : ZoomControl::new(ZOOM_LEVELS.to_vec(), (1.0 * dpi_factor).round()),
+            dpi_factor,
             current_slice : 0,
             tics : 0,
             draw_tool_on : false,
@@ -804,6 +893,9 @@ impl Document for SheetDocument
             scroll_slice_list : false,
             mouse_drawing_texture : false,
             draw_points : Vec::new(),
+            sheet_scroll_x : 0.0,
+            sheet_scroll_y : 0.0,
+            update_sheet_scroll: false,
             dragging : false,
             tilegen : TilesetGenInfo
             {
@@ -947,6 +1039,7 @@ impl Document for SheetDocument
              _project : &mut ProjectConfig, game : &mut GameControl, ui : &imgui::Ui) -> DocumentState
     {
         self.tics = self.tics.wrapping_add(1);
+        self.dpi_factor = game.renderer().current_display_scale_factor();
 
         let mut result = DocumentState::new();
 
@@ -966,11 +1059,12 @@ impl Document for SheetDocument
             });
             ui.menu("Zoom", ||
             {
-                for level in &[0.25, 0.5, 1.0, 2.0, 3.0, 4.0, 6.0, 8.0, 10.0, 12.0, 16.0]
+                for level in &ZOOM_LEVELS
                 {
-                    if imgui::MenuItem::new(format!("{}%", (level * 100.0) as i32)).selected(self.zoom_level == *level).build(ui)
+                    if imgui::MenuItem::new(format!("{}%", (level * 100.0) as i32))
+                        .selected(self.zoom_control.zoom() == *level).build(ui)
                     {
-                        self.zoom_level = *level;
+                        self.apply_zoom(*level);
                     }
                 }
             });
