@@ -1,7 +1,7 @@
 use keeshond::gameloop::GameControl;
 
 use crate::editorgui::{Packages};
-use keeshond_datapack::DataObject;
+use keeshond::datapack::DataObject;
 
 pub mod sheetwizard;
 pub mod costumewizard;
