@@ -3,7 +3,7 @@ use keeshond::gameloop::GameControl;
 use crate::editorgui::Packages;
 use crate::wizard::{Wizard, WizardState};
 
-use keeshond_datapack::DataHandle;
+use keeshond::datapack::DataHandle;
 use keeshond_treats::tilemap::{Tileset, TilesetTile};
 
 pub struct TilesetWizard
